package main

import (
	log "asapo_common/logger"
	"asapo_common/utils"
	"asapo_common/version"
	"asapo_monitoring_server/server"
	"errors"
	"flag"
	"io/ioutil"
	"net/http"
	"net/url"
	"os"
	"strconv"
	"strings"
)

func PrintUsage() {
	log.Fatal("Usage: " + os.Args[0] + " -config <config file>")
}

func loadConfig(configFileName string) (log.Level, server.Settings, error) {
	var settings server.Settings
	if err := utils.ReadJsonFromFile(configFileName, &settings); err != nil {
		return log.FatalLevel, server.Settings{}, err
	}

	if settings.ThisClusterName == "" {
		return log.FatalLevel, server.Settings{}, errors.New("'ThisClusterName' not set")
	}

	if settings.ServerPort == 0 {
		return log.FatalLevel, server.Settings{}, errors.New("'ServerPort' not set")
	}

	if settings.LogLevel == "" {
		return log.FatalLevel, server.Settings{}, errors.New("'LogLevel' not set")
	}

	if settings.InfluxDbUrl == "" {
		return log.FatalLevel, server.Settings{}, errors.New("'InfluxDbUrl' not set")
	}

	if settings.RetentionPolicyTime == "" {
		return log.FatalLevel, server.Settings{}, errors.New("'RetentionPolicyTime' not set")
	}

	if settings.GroupingTime == "" {
		return log.FatalLevel, server.Settings{}, errors.New("'GroupingTime' not set")
	}

	if settings.MaxPoints == 0 {
		return log.FatalLevel, server.Settings{}, errors.New("'MaxPoints' not set")
	}

	if settings.InfluxDbDatabase == "" {
		return log.FatalLevel, server.Settings{}, errors.New("'InfluxDbDatabase' not set")
	}

	logLevel, err := log.LevelFromString(settings.LogLevel)
	if err != nil {
		return log.FatalLevel, server.Settings{}, err
	}

	return logLevel, settings, nil
}

func postTODB(queryPostString string, settings server.Settings) {
	data := url.Values{}
	data.Set("q", queryPostString)

	res, err := http.Post(
		settings.InfluxDbUrl+"/query", "application/x-www-form-urlencoded",
		strings.NewReader(data.Encode()))

	if err != nil {
		log.Fatal(err.Error())
		return
	}
	if res.StatusCode != 200 {
		rawResponse, _ := ioutil.ReadAll(res.Body)
		strResponse := string(rawResponse)

		log.Fatal("failed to do the post, post was: " + queryPostString +
			" and status code was " + strconv.Itoa(res.StatusCode) + ": " + strResponse)
		return
	}
}

func main() {
	var configFileName = flag.String("config", "", "config file path")

	if ret := version.ShowVersion(os.Stdout, "ASAPO Monitoring Server"); ret {
		return
	}

	path, err := os.Getwd()
	println("oath: " + path)

	log.SetSoucre("monitoring server")

	flag.Parse()
	if *configFileName == "" {
		PrintUsage()
	}

	logLevel, settings, err := loadConfig(*configFileName)
	if err != nil {
		log.Fatal(err.Error())
		return
	}
	log.SetLevel(logLevel)

	// InfluxDB 1 fix, create database
	// Does multiple HTTP posts to DB
	// It creates two databases.
	normalDB := settings.InfluxDbDatabase
	avgDB := normalDB + "_avg"
	rpName := "one_day"
	cqName := "avg_values"
	postStrings := []string{
		"CREATE DATABASE " + normalDB,
		"CREATE DATABASE " + avgDB,
		"CREATE RETENTION POLICY " + rpName + " ON " + normalDB +
			" DURATION " + settings.RetentionPolicyTime + " REPLICATION 1 DEFAULT",
		"CREATE CONTINUOUS QUERY " + cqName + " ON " + normalDB + " BEGIN" +
			" SELECT sum(*) INTO " + avgDB +
			"..:MEASUREMENT FROM /.*/ GROUP BY time(" + settings.GroupingTime + "),* END",
	}
	for i := 0; i < len(postStrings); i++ {
		postTODB(postStrings[i], settings)
		log.Debug("Modify DB with command: " + postStrings[i])
	}

	server.Start(settings)
}
