package server

import (
	pb "asapo_common/generated_proto"
	log "asapo_common/logger"
	"context"
	"errors"
	"github.com/influxdata/influxdb-client-go/v2/api"
	"google.golang.org/grpc/codes"
	"google.golang.org/grpc/status"
	"sort"
	"strconv"
	"strings"
	"time"
)

type QueryServer struct {
	pb.UnimplementedAsapoMonitoringQueryServiceServer

	settings        Settings
	dbQueryApi      api.QueryAPI
	switchPeriod    int64
	groupedInterval int64
}

func (s *QueryServer) GetMetadata(ctx context.Context, _ *pb.Empty) (*pb.MetadataResponse, error) {
	result, err := s.dbQueryApi.Query(ctx, ""+
		"tagValues = (bucket, tag, predicate=(r) => true, start=-30d) =>\n"+
		" from(bucket: bucket)\n"+
		"  |> range(start: start)\n"+
		"  |> filter(fn: predicate)\n"+
		"  |> keep(columns: [tag])\n"+
		"  |> group()\n"+
		"  |> distinct(column: tag)"+
		"tagValues(bucket: \""+s.settings.InfluxDbDatabase+"\", tag: \"beamtime\", predicate: (r) => true, start: -30d)")

	// InfluxDB2
	// result, err := s.dbQueryApi.Query(ctx, "import \"influxdata/influxdb/schema\"\n"+
	//	"schema.tagValues(bucket: \""+s.settings.InfluxDbDatabase+"\", tag: \"beamtime\", predicate: (r) => true, start: -30d)")

	if err != nil {
		return nil, err
	}

	response := pb.MetadataResponse{}
	for result.Next() {
		response.AvailableBeamtimes = append(response.AvailableBeamtimes, result.Record().Values()["_value"].(string))
	}
	response.ClusterName = s.settings.ThisClusterName

	return &response, nil
}

func createFilterElement(tagName string, tagValue string) string {
	return "r[\"" + tagName + "\"] == \"" + tagValue + "\""
}

type filterGenerator func(*pb.DataPointsQuery) string

func createFilter(query *pb.DataPointsQuery, generator filterGenerator) string {
	// always present in database: beamtime, source
	filter := " |> filter(fn: (r) => "
	filter += createFilterElement("beamtime", query.BeamtimeFilter)
	if query.SourceFilter != "" {
		filter += " and " + createFilterElement("source", query.SourceFilter)
	}

	filter += generator(query)

	filter += ")"

	return filter
}

func transferSpeedFilterGenerator(query *pb.DataPointsQuery) string {
	filter := ""

	if query.ReceiverFilter != "" {
		filter += " and " + createFilterElement("receiverName", query.ReceiverFilter)
	}
	if query.FromPipelineStepFilter != "" && query.ToPipelineStepFilter != "" {
		if query.PipelineQueryMode == pb.PipelineConnectionQueryMode_ExactPath {
			filter += " and (" +
				// From: receiver
				"(r._measurement == \"" + dbMeasurementFileInput + "\" and " + createFilterElement("pipelineStepId", query.FromPipelineStepFilter) + ")" +
				" or " +
				// To: broker or RDS or FTS
				"((r._measurement == \"" + dbMeasurementBrokerFileRequests + "\" or r._measurement == \"" + dbMeasurementRdsFileRequests + "\" or r._measurement == \"" + dbMeasurementFtsTransfers + "\") and " + createFilterElement("pipelineStepId", query.ToPipelineStepFilter) + ")" +
				")"
		} else if query.PipelineQueryMode == pb.PipelineConnectionQueryMode_JustRelatedToStep {
			// Direction is ignored here. Just check if piplineStepId is different
			if query.FromPipelineStepFilter == query.ToPipelineStepFilter {
				// from and to are the same
				filter += " and " + createFilterElement("pipelineStepId", query.FromPipelineStepFilter)
			} else {
				filter += " and (" +
					createFilterElement("pipelineStepId", query.FromPipelineStepFilter) +
					" or " +
					createFilterElement("pipelineStepId", query.ToPipelineStepFilter)
			}
		}
	}

	return filter
}

func transferRateFilterGenerator(query *pb.DataPointsQuery) string {
	filter := ""

	if query.ReceiverFilter != "" {
		filter += " and " + createFilterElement("receiverName", query.ReceiverFilter)
	}
	if query.FromPipelineStepFilter != "" {
		filter += " and " + createFilterElement("pipelineStepId", query.FromPipelineStepFilter)
	}

	return filter
}

func taskTimeFilterGenerator(query *pb.DataPointsQuery) string {
	filter := ""

	if query.ReceiverFilter != "" {
		filter += " and " + createFilterElement("receiverName", query.ReceiverFilter)
	}
	if query.PipelineQueryMode == pb.PipelineConnectionQueryMode_ExactPath {
		filter += " and (" +
			// From: receiver
			"(r._measurement == \"" + dbMeasurementFileInput + "\" and " + createFilterElement("pipelineStepId", query.FromPipelineStepFilter) + ")" +
			" or " +
			// To: RDS or FTS
			"((r._measurement == \"" + dbMeasurementRdsFileRequests + "\" or r._measurement == \"" + dbMeasurementFtsTransfers + "\") and " + createFilterElement("pipelineStepId", query.ToPipelineStepFilter) + ")" +
			")"
	} else if query.PipelineQueryMode == pb.PipelineConnectionQueryMode_JustRelatedToStep {
		// Direction is ignored here. Just check if piplineStepId is different
		if query.FromPipelineStepFilter == query.ToPipelineStepFilter {
			// from and to are the same
			filter += " and " + createFilterElement("pipelineStepId", query.FromPipelineStepFilter)
		} else {
			filter += " and (" +
				createFilterElement("pipelineStepId", query.FromPipelineStepFilter) +
				" or " +
				createFilterElement("pipelineStepId", query.ToPipelineStepFilter)
		}
	}

	return filter
}

func memoryUsageFilterGenerator(query *pb.DataPointsQuery) string {
	filter := ""

	if query.ReceiverFilter != "" {
		filter += " and " + createFilterElement("receiverName", query.ReceiverFilter)
	}

	return filter
}

// Checks if the string contains any characters that normally should not appear and
// that might interfere with InfluxDb2 Query
func doesStringContainsDangerousElements(input string) bool {
	// ref: https://github.com/influxdata/influxdb-client-js/blob/v1.15.0/packages/core/src/query/flux.ts#L40-L99
	return strings.Contains(input, "\"") ||
		strings.Contains(input, "\n") ||
		strings.Contains(input, "\r") ||
		strings.Contains(input, "\\") ||
		strings.Contains(input, "$") ||
		strings.Contains(input, "{") ||
		strings.Contains(input, "}")
}

func GetRequestPeriods(fromTimestamp uint64, toTimestamp uint64, groupedInterval int64,
	switchPeriod int64, maxPoints uint64) (string, string, int) {
	// Define time ranges for full and reduce DBs and interval between data points

	startTime := strconv.FormatUint(fromTimestamp, 10)
	endTime := strconv.FormatUint(toTimestamp, 10)

	intervalInSec := int((toTimestamp - fromTimestamp) / maxPoints)
	if intervalInSec < 1 {
		intervalInSec = 1
	}

	switchTime := uint64(time.Now().Unix() - switchPeriod)
	timeRangeAve := ""
	timeRange := "start: " + startTime + ", stop: " + endTime
	if fromTimestamp < switchTime {
		intervalInSec = int(groupedInterval)
		if toTimestamp < switchTime {
			timeRangeAve = "start: " + startTime + ", stop: " + endTime
			timeRange = "start: 0, stop: 1 "
		} else {
			timeRangeAve = "start: " + startTime + ", stop: " + strconv.FormatUint(switchTime, 10)
			timeRange = "start: " + strconv.FormatUint(switchTime, 10) + ", stop: " + endTime
		}
	}
	return timeRange, timeRangeAve, intervalInSec
}

func (s *QueryServer) GetDataPoints(ctx context.Context, query *pb.DataPointsQuery) (*pb.DataPointsResponse, error) {

	timeRange, timeRangeAve, intervalInSec := GetRequestPeriods(query.FromTimestamp,
		query.ToTimestamp, s.groupedInterval, s.switchPeriod, s.settings.MaxPoints)
	nEntries := int((query.ToTimestamp - query.FromTimestamp) / uint64(intervalInSec))

	if query.BeamtimeFilter == "" {
		return nil, status.Errorf(codes.InvalidArgument, "Beamtime is required")
	}

	// InfluxDB 2 does not really support input parameters. So the best attempt is to
	// if any from them include '"' or '\' throw an error
	if doesStringContainsDangerousElements(query.BeamtimeFilter) ||
		doesStringContainsDangerousElements(query.FromPipelineStepFilter) ||
		doesStringContainsDangerousElements(query.ToPipelineStepFilter) ||
		doesStringContainsDangerousElements(query.ReceiverFilter) ||
		doesStringContainsDangerousElements(query.SourceFilter) {
		return nil, errors.New("some input characters are not allowed")
	}

	response := pb.DataPointsResponse{}

	transferSpeedFilter := createFilter(query, transferSpeedFilterGenerator)
	transferSpeed, err := queryTransferSpeed(s, ctx, timeRange, timeRangeAve, intervalInSec, transferSpeedFilter, nEntries)
	if err != nil {
		return nil, err
	}
	response.TransferRates = transferSpeed

	fileRateFilter := createFilter(query, transferRateFilterGenerator)
	fileRates, err := queryFileRate(s, ctx, timeRange, timeRangeAve, intervalInSec, fileRateFilter, nEntries)
	if err != nil {
		return nil, err
	}
	response.FileRates = fileRates

	taskTimeFilter := createFilter(query, taskTimeFilterGenerator)
	taskTime, err := queryTaskTime(s, ctx, timeRange, timeRangeAve, intervalInSec, taskTimeFilter, nEntries)
	if err != nil {
		return nil, err
	}
	response.TaskTimes = taskTime

	memoryUsageFilter := createFilter(query, memoryUsageFilterGenerator)
	memoryUsage, err := queryMemoryUsage(s, ctx, timeRange, timeRangeAve, intervalInSec, memoryUsageFilter, nEntries)
	if err != nil {
		return nil, err
	}
	response.MemoryUsages = memoryUsage

	response.StartTimestampInSec = query.FromTimestamp // TODO: Use first timestamp from query result
	response.TimeIntervalInSec = uint32(intervalInSec)
	return &response, nil
}

// Queries DB for transferspeed
func queryfunctionTransferSpeed(s *QueryServer, ctx context.Context,
	timeRange string, bucket string, inset string, intervalInSec int, filter string) ([]*pb.TotalTransferRateDataPoint, error) {

	query := "from(bucket: \"" + bucket + "\")" + // Normal querystring
		" |> range(" + timeRange + ")" +
		" |> filter(fn: (r) => r._measurement == \"" + dbMeasurementFileInput + "\" or r._measurement == \"" + dbMeasurementBrokerFileRequests + "\" or r._measurement == \"" + dbMeasurementRdsFileRequests + "\" or r._measurement == \"" + dbMeasurementFtsTransfers + "\")" +
		" |> filter(fn: (r) => r._field == \"" + inset + "totalInputFileSize\" or r._field == \"" + inset + "totalRequestedFileSize\" or r._field == \"" + inset + "totalRdsOutputFileSize\" or r._field == \"" + inset + "totalFtsTransferredFileSize\")" +
		filter +
		" |> group(columns: [\"_field\"])" +
		" |> aggregateWindow(every: " + strconv.Itoa(intervalInSec) + "s, fn: sum, createEmpty: true)" +
		" |> pivot(rowKey:[\"_time\"], columnKey: [\"_field\"], valueColumn: \"_value\")"
	result, err := s.dbQueryApi.Query(ctx, query)

	if err != nil {
		return nil, err
	}

	var arrayResult []*pb.TotalTransferRateDataPoint
	for result.Next() {
		recvBytes := int64(0)
		sendBytes := int64(0)
		sendRdsBytes := int64(0)
		sendFtsBytes := int64(0)

		if result.Record().Values()[inset+"totalInputFileSize"] != nil {
			recvBytes = result.Record().Values()[inset+"totalInputFileSize"].(int64) * 1000
		}
		if result.Record().Values()[inset+"totalRequestedFileSize"] != nil {
			sendBytes = result.Record().Values()[inset+"totalRequestedFileSize"].(int64)
		}
		if result.Record().Values()[inset+"totalRdsOutputFileSize"] != nil {
			sendRdsBytes = result.Record().Values()[inset+"totalRdsOutputFileSize"].(int64)
		}
		if result.Record().Values()[inset+"totalFtsTransferredFileSize"] != nil {
			sendFtsBytes = result.Record().Values()[inset+"totalFtsTransferredFileSize"].(int64)
		}

		arrayResult = append(arrayResult, &pb.TotalTransferRateDataPoint{
			TotalBytesPerSecRecv:    uint64(recvBytes) / uint64(intervalInSec),
			TotalBytesPerSecSend:    uint64(sendBytes) / uint64(intervalInSec),
			TotalBytesPerSecRdsSend: uint64(sendRdsBytes) / uint64(intervalInSec),
			TotalBytesPerSecFtsSend: uint64(sendFtsBytes) / uint64(intervalInSec),
		})
	}

	if result.Err() != nil {
		return nil, result.Err()
	}
	return arrayResult, nil
}
func queryTransferSpeed(s *QueryServer, ctx context.Context, timeRange string,
	timeRangeAve string, intervalInSec int, filter string, nEntr int) ([]*pb.TotalTransferRateDataPoint, error) {

	arrayResult, err := queryfunctionTransferSpeed(s, ctx, timeRange,
		s.settings.InfluxDbDatabase, "", intervalInSec, filter)
	if err != nil {
		return nil, err
	}

	// Querystring for averaged data
	if timeRangeAve != "" {
		secondArrayResult, err := queryfunctionTransferSpeed(s, ctx, timeRangeAve,
			s.settings.InfluxDbDatabase+averageSuffix, "sum_", intervalInSec, filter)
		if err != nil {
			return nil, err
		}
		if len(secondArrayResult) == 0 {
			secondArrayResult = make([]*pb.TotalTransferRateDataPoint, nEntr-len(arrayResult))
		}
		arrayResult = append(secondArrayResult, arrayResult...)
	}
	return arrayResult, nil
}

// Queries DB for Filerate
func queryfunctionFileRate(s *QueryServer, ctx context.Context, timeRange string,
	bucket string, inset string, intervalInSec int, filter string) ([]*pb.TotalFileRateDataPoint, error) {

	query := "from(bucket: \"" + bucket + "\")" + // Normal querystring
		" |> range(" + timeRange + ")" +
		" |> filter(fn: (r) => r._measurement == \"" + dbMeasurementRdsFileRequests + "\" or r._measurement == \"" + dbMeasurementBrokerFileRequests + "\" or r._measurement == \"" + dbMeasurementFtsTransfers + "\")" +
		" |> filter(fn: (r) => r._field == \"" + inset + "totalRdsHits\" or r._field == \"" + inset + "totalRdsMisses\" or r._field == \"" + inset + "requestedFileCount\" or r._field == \"" + inset + "ftsFileCount\")" +
		filter +
		" |> group(columns: [\"_field\"])" +
		" |> aggregateWindow(every: " + strconv.Itoa(intervalInSec) + "s, fn: sum, createEmpty: true)" +
		" |> pivot(rowKey:[\"_time\"], columnKey: [\"_field\"], valueColumn: \"_value\")"

	rdsResponses, err := s.dbQueryApi.Query(ctx, query)
	if err != nil {
		return nil, err
	}

	var arrayResult []*pb.TotalFileRateDataPoint
	for rdsResponses.Next() {
		totalRequests := uint32(0)
		hit := uint32(0)
		miss := uint32(0)
		fromDisk := uint32(0)

		if rdsResponses.Record().Values()[inset+"requestedFileCount"] != nil {
			totalRequests = uint32(rdsResponses.Record().Values()[inset+"requestedFileCount"].(int64))
		}
		if rdsResponses.Record().Values()[inset+"totalRdsHits"] != nil {
			hit = uint32(rdsResponses.Record().Values()[inset+"totalRdsHits"].(int64))
		}
		if rdsResponses.Record().Values()[inset+"totalRdsMisses"] != nil {
			miss = uint32(rdsResponses.Record().Values()[inset+"totalRdsMisses"].(int64))
		}
		if rdsResponses.Record().Values()[inset+"ftsFileCount"] != nil {
			fromDisk = uint32(rdsResponses.Record().Values()[inset+"ftsFileCount"].(int64))
		}

		arrayResult = append(arrayResult, &pb.TotalFileRateDataPoint{
			TotalRequests: totalRequests,
			CacheMisses:   miss,
			FromCache:     hit,
			FromDisk:      fromDisk,
		})
	}

	if rdsResponses.Err() != nil {
		return nil, rdsResponses.Err()
	}

	return arrayResult, nil
}
func queryFileRate(s *QueryServer, ctx context.Context, timeRange string,
	timeRangeAve string, intervalInSec int, filter string, nEntr int) ([]*pb.TotalFileRateDataPoint, error) {
	arrayResult, err := queryfunctionFileRate(s, ctx, timeRange,
		s.settings.InfluxDbDatabase, "", intervalInSec, filter) // Query newer data
	if err != nil {
		return nil, err
	}

	// Querystring for averaged data
	if timeRangeAve != "" {
		secondArrayResult, err := queryfunctionFileRate(s, ctx, timeRangeAve,
			s.settings.InfluxDbDatabase+averageSuffix, "sum_", intervalInSec, filter)
		if err != nil {
			return nil, err
		}
		if len(secondArrayResult) == 0 {
			secondArrayResult = make([]*pb.TotalFileRateDataPoint, nEntr-len(arrayResult))
		}
		arrayResult = append(secondArrayResult, arrayResult...)
	}
	return arrayResult, nil
}

func queryTaskTime(s *QueryServer, ctx context.Context, timeRange string,
	timeRangeAve string, intervalInSec int, filter string, nEntr int) ([]*pb.TaskTimeDataPoint, error) {

	queryfunction := func(s *QueryServer, ctx context.Context, timeRange string, bucket string,
		inset string, intervalInSec int, filter string) ([]*pb.TaskTimeDataPoint, error) {
		query := "from(bucket: \"" + bucket + "\")" + // Normal querystring
			" |> range(" + timeRange + ")" +
			" |> filter(fn: (r) => r._measurement == \"" + dbMeasurementFileInput + "\" or r._measurement == \"" + dbMeasurementRdsFileRequests + "\")" +
			" |> filter(fn: (r) => r._field == \"" + inset + "avgTransferReceiveTimeUs\" or r._field == \"" + inset + "avgWriteIoTimeUs\" or r._field == \"" + inset + "avgDbTimeUs\" or r._field == \"" + inset + "avgRdsOutputTransferTimeUs\")" +
			filter +
			" |> group(columns: [\"_field\"])" +
			" |> aggregateWindow(every: " + strconv.Itoa(intervalInSec) + "s, fn: mean, createEmpty: true)" +
			" |> pivot(rowKey:[\"_time\"], columnKey: [\"_field\"], valueColumn: \"_value\")"

		var arrayResult []*pb.TaskTimeDataPoint
		result, err := s.dbQueryApi.Query(ctx, query)
		if err != nil {
			return nil, err
		}

		for result.Next() {
			receiveIoTimeUs := float64(0)
			writeToDiskTimeUs := float64(0)
			writeToDatabaseTimeUs := float64(0)
			rdsOutputTransferTimeUs := float64(0)

			if result.Record().Values()[inset+"avgTransferReceiveTimeUs"] != nil {
				receiveIoTimeUs = result.Record().Values()[inset+"avgTransferReceiveTimeUs"].(float64)
			}
			if result.Record().Values()[inset+"avgWriteIoTimeUs"] != nil {
				writeToDiskTimeUs = result.Record().Values()[inset+"avgWriteIoTimeUs"].(float64)
			}
			if result.Record().Values()[inset+"avgDbTimeUs"] != nil {
				writeToDatabaseTimeUs = result.Record().Values()[inset+"avgDbTimeUs"].(float64)
			}
			if result.Record().Values()[inset+"avgRdsOutputTransferTimeUs"] != nil {
				rdsOutputTransferTimeUs = result.Record().Values()[inset+"avgRdsOutputTransferTimeUs"].(float64)
			}

			arrayResult = append(arrayResult, &pb.TaskTimeDataPoint{
				ReceiveIoTimeUs:         uint32(receiveIoTimeUs),
				WriteToDiskTimeUs:       uint32(writeToDiskTimeUs),
				WriteToDatabaseTimeUs:   uint32(writeToDatabaseTimeUs),
				RdsSendToConsumerTimeUs: uint32(rdsOutputTransferTimeUs),
			})
		}

		if result.Err() != nil {
			return nil, result.Err()
		}

		return arrayResult, nil
	}

	arrayResult, err := queryfunction(s, ctx, timeRange,
		s.settings.InfluxDbDatabase, "", intervalInSec, filter) // Query newer data
	if err != nil {
		return nil, err
	}

	// Querystring for averaged data
	if timeRangeAve != "" {
		secondArrayResult, err := queryfunction(s, ctx, timeRangeAve,
			s.settings.InfluxDbDatabase+averageSuffix, "sum_", intervalInSec, filter)
		if err != nil {
			return nil, err
		}
		if len(secondArrayResult) == 0 {
			secondArrayResult = make([]*pb.TaskTimeDataPoint, nEntr-len(arrayResult))
		}
		arrayResult = append(secondArrayResult, arrayResult...)
	}
	return arrayResult, nil
}

func queryMemoryUsage(s *QueryServer, ctx context.Context, timeRange string,
	timeRangeAve string, intervalInSec int, filter string, nEntr int) ([]*pb.RdsMemoryUsageDataPoint, error) {

	queryfunction := func(s *QueryServer, ctx context.Context, timeRange string, bucket string,
		inset string, intervalInSec int, filter string) ([]*pb.RdsMemoryUsageDataPoint, error) {
		var arrayResult []*pb.RdsMemoryUsageDataPoint
		// Querystrings
		query := "from(bucket: \"" + s.settings.InfluxDbDatabase + "\")" + // Normal querystring
			" |> range(" + timeRange + ")" +
			" |> filter(fn: (r) => r._measurement == \"" + dbMeasurementRdsCacheMemoryUsage + "\")" +
			" |> filter(fn: (r) => r._field == \"rdsCacheUsedBytes\")" +
			filter +
			" |> group(columns: [\"_field\"])" +
			" |> aggregateWindow(every: " + strconv.Itoa(intervalInSec) + "s, fn: sum, createEmpty: true)" +
			" |> pivot(rowKey:[\"_time\"], columnKey: [\"_field\"], valueColumn: \"_value\")"
		result, err := s.dbQueryApi.Query(ctx, query)
		if err != nil {
			return nil, err
		}
		for result.Next() {
			usedBytes := int64(0)

			if result.Record().Values()[inset+"rdsCacheUsedBytes"] != nil {
				usedBytes = result.Record().Values()[inset+"rdsCacheUsedBytes"].(int64)
			}

			arrayResult = append(arrayResult, &pb.RdsMemoryUsageDataPoint{
				TotalUsedMemory: uint64(usedBytes),
			})
		}
		if result.Err() != nil {
			return nil, result.Err()
		}

		return arrayResult, nil
	}

	arrayResult, err := queryfunction(s, ctx, timeRange,
		s.settings.InfluxDbDatabase, "", intervalInSec, filter) // Query for newer data
	if err != nil {
		return nil, err
	}

	// Querystring for averaged data
	if timeRangeAve != "" {
		secondArrayResult, err := queryfunction(s, ctx, timeRangeAve,
			s.settings.InfluxDbDatabase+averageSuffix, "sum_", intervalInSec, filter)
		if err != nil {
			return nil, err
		}
		if len(secondArrayResult) == 0 {
			secondArrayResult = make([]*pb.RdsMemoryUsageDataPoint, nEntr-len(arrayResult))
		}
		arrayResult = append(secondArrayResult, arrayResult...)
	}

	return arrayResult, nil
}

func (s *QueryServer) GetTopology(ctx context.Context, query *pb.ToplogyQuery) (*pb.TopologyResponse, error) {
	if doesStringContainsDangerousElements(query.BeamtimeFilter) {
		return nil, errors.New("some input characters are not allowed")
	}

	startTime := strconv.FormatUint(query.FromTimestamp, 10)
	endTime := strconv.FormatUint(query.ToTimestamp, 10)

	result, err := s.dbQueryApi.Query(ctx, "from(bucket: \""+s.settings.InfluxDbDatabase+"\")"+
		" |> range(start: "+startTime+", stop: "+endTime+")"+
		" |> filter(fn: (r) => r._measurement == \""+dbMeasurementFileInput+"\" or r._measurement == \""+dbMeasurementBrokerFileRequests+"\")"+
		" |> filter(fn: (r) => "+createFilterElement("beamtime", query.BeamtimeFilter)+")"+
		" |> keep(columns: [\"receiverName\", \"brokerName\", \"pipelineStepId\", \"source\", \"producerInstanceId\", \"consumerInstanceId\"])"+
		" |> group()",
	)

	if err != nil {
		return nil, err
	}

	type PipelineStep struct { // these maps here are replacements for HashSets
		producesSourceId map[string] /*produces sourceId*/ bool
		consumesSourceId map[string] /*consumers sourceId*/ bool

		producerInstances map[string] /*produces instanceId*/ bool
		consumerInstances map[string] /*consumers instanceId*/ bool

		involvedReceivers map[string] /*receiver id*/ bool
	}

	pipelineSteps := map[string] /*pipelineStepId*/ PipelineStep{}

	getOrCreateStep := func(stepId string) PipelineStep {
		if _, containsKey := pipelineSteps[stepId]; !containsKey {
			pipelineSteps[stepId] = PipelineStep{
				producesSourceId:  map[string] /*produces sourceId*/ bool{},
				consumesSourceId:  map[string] /*consumers sourceId*/ bool{},
				producerInstances: map[string] /*produces instanceId*/ bool{},
				consumerInstances: map[string] /*consumers instanceId*/ bool{},
				involvedReceivers: map[string] /*receiver id*/ bool{},
			}
		}
		return pipelineSteps[stepId]
	}

	for result.Next() {
		if result.Record().Values()["receiverName"] != nil { // data is coming from receiver => means it must be a producer
			stepId, ok := result.Record().Values()["pipelineStepId"].(string)
			if !ok {
				stepId = "defaultStep"
			}
			source := result.Record().Values()["source"].(string)
			producerInstanceId := result.Record().Values()["producerInstanceId"].(string)
			var step = getOrCreateStep(stepId)
			step.producesSourceId[source] = true
			step.producerInstances[producerInstanceId] = true
			var receiverName = result.Record().Values()["receiverName"].(string)
			step.involvedReceivers[receiverName] = true
		} else if result.Record().Values()["brokerName"] != nil { // data is coming from broker => means it must be a consumer
			stepId := result.Record().Values()["pipelineStepId"].(string)
			source := result.Record().Values()["source"].(string)
			consumerInstanceId := result.Record().Values()["consumerInstanceId"].(string)
			var step = getOrCreateStep(stepId)
			step.consumesSourceId[source] = true
			step.consumerInstances[consumerInstanceId] = true
		} else {
			log.Debug("Got an entry without receiverName or brokerName")
		}
	}

	type PipelineLevelStepInfo struct {
		stepId string
	}

	type PipelineEdgeInfo struct {
		fromStepId        string
		toStepId          string
		sourceId          string
		involvedReceivers map[string] /*receiver id*/ bool
	}

	var levels [] /*levelDepth*/ [] /*verticalIdx*/ PipelineLevelStepInfo
	var edges []PipelineEdgeInfo
	var currentLevel = -1
	var unkStep = "Unknown"
	var addUnknownStep = false
	unknownSources := map[string] /*sourceId*/ bool{}
	// Loop over all consumers of all pipeline steps
	for inStepId := range pipelineSteps {
		for requiredSourceId := range pipelineSteps[inStepId].consumesSourceId {
			// Search for existing producer for given consumer
			// Create edge ifprodures is found
			var producerFound = false
			for outStepId := range pipelineSteps {
				if pipelineSteps[outStepId].producesSourceId[requiredSourceId] {
					edges = append(edges, PipelineEdgeInfo{
						fromStepId:        outStepId,
						toStepId:          inStepId,
						sourceId:          requiredSourceId,
						involvedReceivers: pipelineSteps[outStepId].involvedReceivers,
					})
					producerFound = true
				}
			}
			// Create edge with Unknown pipeline step is not found
			if !producerFound {
				addUnknownStep = true
				unknownSources[requiredSourceId] = true
				edges = append(edges, PipelineEdgeInfo{
					fromStepId:        unkStep,
					toStepId:          inStepId,
					sourceId:          requiredSourceId,
					involvedReceivers: nil,
				})
			}
		}
	}
	if addUnknownStep {
		pipelineSteps[unkStep] = PipelineStep{
			producesSourceId:  nil,
			consumesSourceId:  nil,
			producerInstances: nil,
			consumerInstances: nil,
			involvedReceivers: nil,
		}
	}

	// Define node dependencies
	remainingStepIds := map[string] /*pipelineStepId*/ bool{}
	availableSources := map[string] /*sourceId*/ bool{}
	for stepId, _ := range pipelineSteps {
		remainingStepIds[stepId] = true
	}

	// limit of 10 levels is given to avoid stuck in case of bug
	for len(remainingStepIds) > 0 && currentLevel < 10 {
		currentLevel++
		levels = append(levels, []PipelineLevelStepInfo{})
		currentSources := map[string]bool{}
		for stepId := range remainingStepIds {

			// If not, check if at this level all sources are already available
			var allAvailable = true
			for consumerSource, _ := range pipelineSteps[stepId].consumesSourceId {
				if !availableSources[consumerSource] && !pipelineSteps[stepId].producesSourceId[consumerSource] {
					allAvailable = false
				}
			}
			if allAvailable {

				levels[currentLevel] = append(levels[currentLevel], PipelineLevelStepInfo{
					stepId: stepId,
				})
				// Add all receiver of given pipeline step to current available sources
				for sourceId := range pipelineSteps[stepId].producesSourceId {
					currentSources[sourceId] = true
				}
				delete(remainingStepIds, stepId)
			}
		}
		if currentLevel == 0 {
			for source, _ := range unknownSources {
				availableSources[source] = true
			}
		}
		for source, _ := range currentSources {
			availableSources[source] = true
		}
	}

	// Response building
	var response pb.TopologyResponse
	for levelIndex, level := range levels {
		for _, node := range level {
			var producers []string = nil
			var consumers []string = nil

			for producerInstanceId := range pipelineSteps[node.stepId].producerInstances {
				producers = append(producers, producerInstanceId)
			}
			for consumerInstanceId := range pipelineSteps[node.stepId].consumerInstances {
				consumers = append(consumers, consumerInstanceId)
			}
			response.Nodes = append(response.Nodes, &pb.TopologyResponseNode{
				NodeId:            node.stepId,
				Level:             uint32(levelIndex),
				ProducerInstances: producers,
				ConsumerInstances: consumers,
			})
		}
	}

	for _, edge := range edges {
		newEdge := pb.TopologyResponseEdge{
			FromNodeId:        edge.fromStepId,
			ToNodeId:          edge.toStepId,
			SourceName:        edge.sourceId,
			InvolvedReceivers: nil,
		}

		for receiverId := range edge.involvedReceivers {
			newEdge.InvolvedReceivers = append(newEdge.InvolvedReceivers, receiverId)
		}

		response.Edges = append(response.Edges, &newEdge)
	}

	return &response, nil
}

func (s *QueryServer) GetGroupDelay(ctx context.Context, query *pb.GroupDelayQuery) (*pb.GroupDelayResponse, error) {

	if query.BeamtimeFilter == "" {
		return nil, status.Errorf(codes.InvalidArgument, "Beamtime is required")
	}

	if doesStringContainsDangerousElements(query.BeamtimeFilter) {
		return nil, errors.New("some input characters are not allowed")
	}

	startTime := strconv.FormatUint(query.FromTimestamp, 10)
	endTime := strconv.FormatUint(query.ToTimestamp, 10)

	result, err := s.dbQueryApi.Query(ctx, "from(bucket: \""+s.settings.InfluxDbDatabase+"\")"+
		" |> range(start: "+startTime+", stop: "+endTime+")"+
		" |> filter(fn: (r) => r._measurement == \""+dbMeasurementBrokerFileRequests+"\""+" )"+
		" |> filter(fn: (r) => r.beamtime == \""+query.BeamtimeFilter+"\" and r._field == \"delayMs\")"+
		" |> group(columns: [\"groupId\"])"+
		" |> last()"+
		" |> keep(columns: [\"groupId\", \"_value\", \"pipelineStepId\"])"+
		" |> group()",
	)
	if err != nil {
		return nil, err
	}

	response := pb.GroupDelayResponse{}
	for result.Next() {
		if result.Record().Value() != nil && result.Record().Values()["pipelineStepId"] != nil {
			groupId := result.Record().Values()["groupId"]
			if groupId == nil {
				groupId = ""
			}
			item := pb.GroupDelayResponseItem{
				GroupId:      groupId.(string),
				DelayMs:      (uint32)(result.Record().Value().(int64)),
				PipelineStep: result.Record().Values()["pipelineStepId"].(string),
			}
			response.GroupIds = append(response.GroupIds, &item)
		}
	}

	sort.SliceStable(response.GroupIds, func(i, j int) bool {
		return response.GroupIds[i].DelayMs > response.GroupIds[j].DelayMs
	})

	return &response, nil
}
